// Functions contain dynamically generated IL code. They are the end
// result of the chain Tokenizer -> Parser -> ParseTree -> Function.

using System.Collections.Generic;
using System.Reflection.Emit;



class Function
{
    // compile a parse tree into executable code
    public Function(ParseTree parseTree, string[] locals)
    {
        if (locals == null)
            locals = new string[0];

        DynamicMethod body = Compiler.CompileFunction(parseTree, locals);

        switch (locals.Length) {

            case 0:  mBody0 = (Body0)body.CreateDelegate(typeof(Body0)); break;
            case 1:  mBody1 = (Body1)body.CreateDelegate(typeof(Body1)); break;
            case 2:  mBody2 = (Body2)body.CreateDelegate(typeof(Body2)); break;
            case 3:  mBody3 = (Body3)body.CreateDelegate(typeof(Body3)); break;
            case 4:  mBody4 = (Body4)body.CreateDelegate(typeof(Body4)); break;

            default: mBodyN = (BodyN)body.CreateDelegate(typeof(BodyN)); break;
        }

        SourceCode = parseTree.SourceCode;
    }



    // remember the source text, in case we want to display or save it later on
    public readonly string SourceCode;



    // evaluate functions with different numbers of arguments
    public Complex Evaluate(Globals globals)                                { return mBody0(MaxRecursion, globals); }
    public Complex Evaluate(Globals globals, Complex arg1)                  { return mBody1(MaxRecursion, globals, arg1); }
    public Complex Evaluate(Globals globals, Complex arg1, Complex arg2)    { return mBody2(MaxRecursion, globals, arg1, arg2); }



    // limit how many times we can recurse
    const int MaxRecursion = 1024;



    // optimized body prototypes for different numbers of arguments
    delegate Complex Body0(int recursion, Globals globals);
    delegate Complex Body1(int recursion, Globals globals, Complex arg1);
    delegate Complex Body2(int recursion, Globals globals, Complex arg1, Complex arg2);
    delegate Complex Body3(int recursion, Globals globals, Complex arg1, Complex arg2, Complex arg3);
    delegate Complex Body4(int recursion, Globals globals, Complex arg1, Complex arg2, Complex arg3, Complex arg4);
    delegate Complex BodyN(int recursion, Globals globals, Complex[] args);

    Body0 mBody0 = delegate { ArgCountError(); return 0; };
    Body1 mBody1 = delegate { ArgCountError(); return 0; };
    Body2 mBody2 = delegate { ArgCountError(); return 0; };
    Body3 mBody3 = delegate { ArgCountError(); return 0; };
    Body4 mBody4 = delegate { ArgCountError(); return 0; };
    BodyN mBodyN = delegate { ArgCountError(); return 0; };

    public const int MaxArgs = 4;



    // complain if the number of arguments is wrong
    static void ArgCountError()
    {
        throw new ExpressionException("Wrong number of arguments for user defined function.");
    }



    // check if an assignment expression is actually a function declaration
    public static bool Munge(ParseTree parseTree, out string name, out string[] locals, out ParseTree body)
    {
        name = null;
        locals = null;
        body = null;

        // we want something like x=y or f(x)=y
        ParseTreeOperator assign = parseTree as ParseTreeOperator;

        if ((assign == null) ||
            (assign.Operator != OpTable.Assign) ||
            (assign.Args.Length != 2))
            return false;

        ParseTreeFunction function = assign.Args[0] as ParseTreeFunction;

        if (function == null)
            return false;

        // pull out a list of argument names
        List<string> arglist = new List<string>();

        foreach (ParseTree a in function.Args) {
            ParseTreeFunction arg = a as ParseTreeFunction;

            if ((arg == null) || (arg.Args.Length > 0))
                return false;

            arglist.Add(arg.Name);
        }

        // yay!
        name = function.Name;
        locals = arglist.ToArray();
        body = assign.Args[1];

        // propagate the source text
        body.SourceCode = parseTree.SourceCode;

        return true;
    }
}

