// Read the Tokenizer, write the ParseTree. Thus the cookie crumbles.

using System;
using System.Collections.Generic;



class Parser
{
	public Parser(TokenPeeker tokenizer, bool nested)
	{
		int start = tokenizer.Pos;
		int end = start;

		// shift/reduce loop
		while (!DoneParsing(tokenizer, nested)) {
			object token = tokenizer.Read();

			Operator op = token as Operator;

			if (op != null)
				PushOperator(op);
			else if (token != null)
				PushValue(token, tokenizer);
			else
				throw ParseError();

			end = tokenizer.Pos;
		}

		// flush the stack
		if (mCurrent == null)
			throw ParseError();

		PushOperator(OpTable.Terminator);

		if ((mOperators.Count != 1) || (mValues.Count != 1))
			throw ParseError();

		// remember what part of the input string we used
		Result.SourceCode = tokenizer.Input.Substring(start, end-start).Trim();
	}



	// push a value onto the stack
	void PushValue(object value, TokenPeeker tokenizer)
	{
		if (mCurrent != null)
			throw ParseError();

		string str = value as string;

		if (str != null) {
			// some kind of function call or variable lookup
			ParseTree[] args = ParseArguments(tokenizer);

			Operator op;

			if (OpTable.Functions.TryGetValue(str, out op))
				mCurrent = new ParseTreeOperator(op, args);
			else
				mCurrent = new ParseTreeFunction(str, args);
		}
		else {
			// numeric constant
			mCurrent = new ParseTreeConstant((double)value);
		}
	}



	// push an operator onto the stack
	void PushOperator(Operator op)
	{
		// bodge to handle unary negation versus binary subtraction
		if ((op == OpTable.Subtract) && (mCurrent == null))
			op = OpTable.UnaryNegate;

		// reduce the precedence stack
		if (op != OpTable.OpenBrace) {
			Ops.Precedence precedence = op.Precedence;

			while (mOperators.Count > 0) {
				// compare precedence with the operator on top of the stack
				Operator stack = mOperators.Peek();

				Ops.Precedence stackPrecedence = stack.Precedence;

				if (stack.Arity != 1) {
					if (op.Arity == 1)
						break;

					if (!stack.RightAssociative)
						stackPrecedence++;
				}

				if (precedence >= stackPrecedence)
					break;

				// pop from the stack
				mOperators.Pop();

				ParseTree stackValue = mValues.Pop();

				switch (stack.Arity) {

					case 1:
						// unary operator
						if ((mCurrent == null) || (stackValue != null))
							throw ParseError();

						mCurrent = new ParseTreeOperator(stack, new ParseTree[] { mCurrent } );
						break;

					case 2:
						// binary operator
						if ((mCurrent == null) || (stackValue == null))
							throw ParseError();

						mCurrent = BinaryOrTernary(stack, stackValue, mCurrent);
						break;

					default:
						// match open and close braces
						if ((stack != OpTable.OpenBrace) || (stackValue != null))
							throw ParseError();

						if (op == OpTable.CloseBrace) {
							if (mCurrent == null)
								throw ParseError();

							return;
						}
						else
							break;
				}
			}
		}

		if (op == OpTable.CloseBrace) {
			// swallow close braces, but not too many
			if (mOperators.Count <= 0)
				throw ParseError();
		}
		else {
			// push onto the stack
			mOperators.Push(op);
			mValues.Push(mCurrent);
			mCurrent = null;
		}
	}



	// decide whether this is a binary or ternary operator
	static ParseTree BinaryOrTernary(Operator op, ParseTree x, ParseTree y)
	{
		if (op == OpTable.QuestionMark) {
			ParseTreeOperator colon = y as ParseTreeOperator;

			if ((colon != null) && (colon.Operator == OpTable.Colon))
				return new ParseTreeOperator(OpTable.Ternary, new ParseTree[] { x, colon.Args[0], colon.Args[1] } );
		}

		return new ParseTreeOperator(op, new ParseTree[] { x, y } );
	}



	// parse the arguments of a function call
	static ParseTree[] ParseArguments(TokenPeeker tokenizer)
	{
		List<ParseTree> args = new List<ParseTree>();

		if (tokenizer.Peek() == OpTable.OpenBrace) {
			tokenizer.Read();

			while (tokenizer.Peek() != OpTable.CloseBrace)
				args.Add(new Parser(tokenizer, true).Result);

			tokenizer.Read();
		}

		return args.ToArray();
	}



	// decide if we've reached the end of the expression
	bool DoneParsing(TokenPeeker tokenizer, bool nested)
	{
		// commas always terminate
		if (tokenizer.Peek() as string == ",") {
			tokenizer.Read();
			return true;
		}

		if (nested) {
			// parsing x or y from something like f(x, y(z))
			if (tokenizer.Peek() == OpTable.CloseBrace) {
				foreach (Operator op in mOperators) {
					if (op == OpTable.OpenBrace)
						return false;
				}

				return true;
			}
			else
				return false;
		}
		else {
			// parsing the top level expression
			return (tokenizer.Peek() == null);
		}
	}



	// complain when things go wrong
	static Exception ParseError()
	{
		return new ExpressionException("Invalid expression.");
	}



	// token stacks
	Stack<Operator>  mOperators = new Stack<Operator>();
	Stack<ParseTree> mValues    = new Stack<ParseTree>();

	ParseTree mCurrent;



	// expose the resulting parse tree to the outside world
	public ParseTree Result
	{
		get { return mValues.Peek(); }
	}
}

