// Splits a messy input string into nice tidy tokens.

using System.Collections;
using System.Globalization;
using System.Text;



class Tokenizer : IPeekable
{
	// start tokenizing a string
	public Tokenizer(string input, IDictionary operators)
	{
		mOperators = operators;
		mInput = input;
	}



	// token reader
	public object Read()
	{
		while (char.IsWhiteSpace(Peek()))
			Next();

		char c = Peek();

		if (c == EOF)
			return null;

		// numbers
		if (char.IsDigit(c) || (c == '.'))
			return ReadNumber();

		// barewords
		if (char.IsLetter(c) || (c == '_'))
			return ReadBareword();

		// quoted strings
		if ((c == '"') || (c == '\''))
			return ReadQuoted();

		// a special operator?
		object op = ReadOperator();

		if (op != null)
			return op;

		// unknown single character
		return Next().ToString();
	}



	// read a numeric constant
	double ReadNumber()
	{
		string s = "";

		if (Peek() == '0') {
			s += Next();

			switch (Peek()) {
				case 'x': Next(); return ReadHex();	
				case 'b': Next(); return ReadBinary();	
			}
		}

		return ReadDecimal(s);
	}



	// read a decimal floating point constant
	double ReadDecimal(string s)
	{
		while (char.IsDigit(Peek()) || (Peek() == '.') || (Peek() == 'e')) {
			if (Peek() == 'e') {
				s += Next();

				if (Peek() == '-')
					s += Next();
			}
			else
				s += Next();
		}

		return double.Parse(s, CultureInfo.InvariantCulture);
	}



	// convert a hex character to a numeric value
	static int Hex(char c)
	{
		if (char.IsDigit(c))
			return c - '0';

		c = char.ToUpper(c);

		if ((c >= 'A') && (c <= 'F'))
			return 0xA + c - 'A';

		return -1;
	}



	// read a hexadecimal constant
	double ReadHex()
	{
		ulong val = 0;

		while (Hex(Peek()) >= 0) {
			val <<= 4;
			val |= (uint)Hex(Next());
		}

		return val;
	}



	// read a binary constant
	double ReadBinary()
	{
		ulong val = 0;

		while ((Peek() == '0') || (Peek() == '1')) {
			val <<= 1;
			val |= (uint)(Next() - '0');
		}

		return val;
	}



	// read a bareword
	string ReadBareword()
	{
		StringBuilder s = new StringBuilder();

		while (char.IsLetterOrDigit(Peek()) || (Peek() == '_'))
			s.Append(Next());

		return string.Intern(s.ToString());
	}



	// read a quoted string
	string ReadQuoted()
	{
		char quote = Next();
		char c = Next();

		// empty string (two quotes together)
		if (quote == c)
			return quote.ToString() + quote.ToString();

		// otherwise read up to the closing quote
		StringBuilder s = new StringBuilder();

		while ((c != quote) && (c != EOF)) {
			s.Append(c);
			c = Next();
		}

		return s.ToString();
	}



	// check if this could be an operator
	bool CouldBeOperator(string s)
	{
		foreach (string t in mOperators.Keys) {
			if (t.StartsWith(s))
				return true;
		}

		return false;
	}



	// read an operator
	object ReadOperator()
	{
		string s = "";

		while (CouldBeOperator(s + Peek()))
			s += Next();

		return mOperators[s];
	}



	// arbitrary operator list
	IDictionary mOperators;



	// input string and read position
	string mInput;

	int mPos;



	// end marker
	const char EOF = '\0';



	// helpers for reading character input
	char Peek()
	{
		return (mPos < mInput.Length) ? mInput[mPos] : EOF;
	}

	char Next()
	{
		return (mPos < mInput.Length) ? mInput[mPos++] : EOF;
	}



	// expose the input string and read position
	public string Input
	{
		get { return mInput; }
	}

	public int Pos
	{
		get { return mPos; }
	}
}

