// A curve wraps a function for display by the plot window.

using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Collections.Generic;
using System.Runtime.InteropServices;



// base class for the various types of curve
abstract class Curve
{
	protected Curve(AxisRange autorange)
	{
		AutoRange = autorange;
	}



	// which axis ranges do we know how to compute?
	public enum AxisRange
	{
		None = 0,
		X    = 1,
		Y    = 2,
		XY   = 3,
	}

	public readonly AxisRange AutoRange;



	// expose the source expression
	public abstract string SourceCode { get; }



	// evaluate the curve, returning whatever kind of data the Paint() method wants
	abstract public object Evaluate(Globals globals, int w, int h, Pair<Range> range, Color color, ProgressHandler progress);



	// hook for updating the display if Evaluate() is taking a long time
	public delegate void ProgressHandler(Curve curve, Object data, float pos);



	// helper for evaluating a range of values along one axis of a curve
	static protected List<double> Evaluate(Globals globals, Function function, int n, ref Range srcRange, ref Range destRange)
	{
		// lock down the source range
		if (srcRange == null)
			srcRange = new Range();

		srcRange.Lock();

		// points on the curve to find (Luciano Berio, 1974)
		List<double> points = new List<double>(n);

		foreach (double pos in srcRange.Enumerate(n))
			points.Add(function.Evaluate(globals, pos).Real);

		// update the destination range as appropriate
		if (destRange == null)
			destRange = MeasureRange(points);
		else if (!destRange.Locked)
			destRange.Add(MeasureRange(points));

		return points;
	}



	// estimate what might be a good range for displaying one axis of a curve
	static Range MeasureRange(List<double> data)
	{
		const int    Outliers = 32;
		const double Padding  = 0.15;
		const double Minimum  = 0.1;

		// ignore degenerate values
		data = data.FindAll(IsFinite);

		if (data.Count == 0)
			return new Range();

		// discard the extremities
		double[] median = Util.Sort(data);

		double min = median[median.Length / Outliers];
		double max = median[median.Length * (Outliers-1) / Outliers];

		// pad back up to a good size
		double padding = (max - min) * Padding;

		padding = Math.Max(padding, Minimum);

		return new Range(min-padding, max+padding);
	}



	// default way of rendering a curve to the screen
	virtual public void Paint(Graphics gfx, Size size, Matrix transform, Color color, object data, object recalc, float recalcPos)
	{
		List<PointF> pointlist = data as List<PointF>;

		if (pointlist == null)
			return;

		// apply the range transform
		PointF[] points = pointlist.ToArray();

		transform.TransformPoints(points);

		// don't try to draw anything too silly
		for (int i=0; i<points.Length; i++) {
			points[i].X = ClampToScreen(points[i].X);
			points[i].Y = ClampToScreen(points[i].Y);
		}

		int pos = 0;

		while (pos < points.Length) {
			// skip bogus entries
			pos = Array.FindIndex(points, pos, IsFinite);

			if (pos < 0)
				break;

			// find a connected range of points
			int end = pos+1;

			while ((end < points.Length) && (IsFinite(points[end])) && (AreConnected(points[end], points[end-1], size)))
				end++;

			// draw the range
			int count = end - pos;

			if (count > 1) {
				using (Pen pen = new Pen(color)) {
					if ((pos == 0) && (end == points.Length)) {
						// if possible, render directly from the source array
						gfx.DrawLines(pen, points);
					}
					else {
						// otherwise we have to copy out a subregion
						PointF[] tmp = new PointF[count];

						Array.Copy(points, pos, tmp, 0, count);

						gfx.DrawLines(pen, tmp);
					}
				}
			}

			pos = end;
		}
	}



	// helper for constructing point objects
	static protected PointF MakePoint(double x, double y)
	{
		return new PointF((float)x, (float)y);
	}



	// test whether a number is finite
	static bool IsFinite(float f)
	{
		return !float.IsInfinity(f) && !float.IsNaN(f);
	}

	static bool IsFinite(double d)
	{
		return IsFinite((float)d);
	}

	static bool IsFinite(PointF p)
	{
		return IsFinite(p.X) && IsFinite(p.Y);
	}



	// make a coordinate be reasonable for rendering to the screen
	static float ClampToScreen(float f)
	{
		const float threshold = 4096;

		if (float.IsNegativeInfinity(f) || (f < -threshold))
			return -threshold;

		if (float.IsPositiveInfinity(f) || (f > threshold))
			return threshold;

		return f;
	}



	// check if two points are close enough to be worth connecting
	static bool AreConnected(PointF a, PointF b, Size size)
	{
		// threshold depends on whether we are crossing a sign boundary
		float w = size.Width/2;
		float h = size.Height/2;

		float threshold = (w+h)/2;

		if (((a.X-w < 0) == (b.X-w < 0)) && ((a.Y-h < 0) == (b.Y-h < 0)))
			threshold *= 1024;

		// son of the squaw
		float dx = a.X - b.X;
		float dy = a.Y - b.Y;

		float d = dx*dx + dy*dy;

		return (d < threshold*threshold);
	}
}



// base class for curves that display one single function
abstract class SingleCurve : Curve
{
	protected SingleCurve(ParseTree parseTree, string[] args, AxisRange autorange)
		: base(autorange)
	{
		mFunction = new Function(parseTree, args);
	}


	protected Function mFunction;


	override public string SourceCode
	{
		get { return mFunction.SourceCode; }
	}
}



// a curve of the form y=f(x)
class YCurve : SingleCurve
{
	public YCurve(ParseTree parseTree)
		: base(parseTree, new string[] { "x" }, AxisRange.Y)
		{ }


	override public object Evaluate(Globals globals, int w, int h, Pair<Range> range, Color color, ProgressHandler progress)
	{
		List<double> points = Evaluate(globals, mFunction, w, ref range.X, ref range.Y);

		return Util.Zip<PointF, double, double>(range.X.Enumerate(w), points, MakePoint);
	}
}



// a curve of the form x=f(y)
class XCurve : SingleCurve
{
	public XCurve(ParseTree parseTree)
		: base(parseTree, new string[] { "y" }, AxisRange.X)
		{ }


	override public object Evaluate(Globals globals, int w, int h, Pair<Range> range, Color color, ProgressHandler progress)
	{
		List<double> points = Evaluate(globals, mFunction, h, ref range.Y, ref range.X);

		return Util.Zip<PointF, double, double>(points, range.Y.Enumerate(h), MakePoint);
	}
}



// a curve of the form x=f(t), y=f(t)
class ParametricCurve : Curve
{
	public ParametricCurve(ParseTree x, ParseTree y, Pair<Function> range)
		: base(AxisRange.XY)
	{
		string[] t = { "t" };

		mFunctionX = new Function(x, t);
		mFunctionY = new Function(y, t);

		mRange = range;
	}



	override public object Evaluate(Globals globals, int w, int h, Pair<Range> range, Color color, ProgressHandler progress)
	{
		// evaluate the parameter range
		Range t = new Range(mRange.X.Evaluate(globals).Real,
							mRange.Y.Evaluate(globals).Real);

		// evaluate both functions
		List<double> x = Evaluate(globals, mFunctionX, NumSamples, ref t, ref range.X);
		List<double> y = Evaluate(globals, mFunctionY, NumSamples, ref t, ref range.Y);

		// interleave the results
		return Util.Zip<PointF, double, double>(x, y, MakePoint);
	}



	// a pair of functions, plus a range for the t value
	Function mFunctionX;
	Function mFunctionY;

	Pair<Function> mRange;

	const int NumSamples = 2048;



	// combine the source of both functions
	override public string SourceCode
	{
		get { return mFunctionX.SourceCode + ", " + mFunctionY.SourceCode; }
	}
}



// a curve of the form s=f(x,y) colors pixels according to an expression set
class SetCurve : SingleCurve
{
	public SetCurve(ParseTree parseTree)
		: base(parseTree, new string[] { "x", "y" }, AxisRange.None)
		{ }



	override public object Evaluate(Globals globals, int w, int h, Pair<Range> range, Color color, ProgressHandler progress)
	{
		// make sure we have a valid range
		for (int i=0; i<2; i++) {
			if (range[i] == null)
				range[i] = new Range();

			range[i].Lock();
		}

		// fill an array of color values
		int rgb = color.ToArgb() & 0xFFFFFF;

		int[] data = new int[w*h];

		int pos = 0;

		Complex[] args = new Complex[2];

		int start = Environment.TickCount;

		// evaluate the set
		foreach (double y in range.Y.Enumerate(h)) {
			foreach (double x in range.X.Enumerate(w)) {
				int a = (int)(mFunction.Evaluate(globals, x, y).Real * 255);

				a = Util.Clamp(0, a, 255);

				data[pos++] = rgb | (a << 24);
			}

			// update the display if this is taking a long time
			if ((progress != null) && (Environment.TickCount > start+50)) {
				int p = pos/w;
				progress(this, CreateBitmap(data, w, p), (float)p/h);
				start = Environment.TickCount;
			}
		}

		return CreateBitmap(data, w, h);
	}



	// turn an array of colors into a UI bitmap
	protected virtual Bitmap CreateBitmap(int[] data, int w, int h)
	{
		Bitmap bitmap = new Bitmap(w, h);

		BitmapData locked = bitmap.LockBits(new Rectangle(Point.Empty, bitmap.Size), ImageLockMode.WriteOnly, bitmap.PixelFormat);

		Marshal.Copy(data, 0, locked.Scan0, w*h);

		bitmap.UnlockBits(locked);

		return bitmap;
	}



	// render the curve to the screen
	override public void Paint(Graphics gfx, Size size, Matrix transform, Color color, object data, object recalc, float pos)
	{
		Image img1 = data as Image;
		Image img2 = recalc as Image;

		// draw the main image
		if (img1 != null) {
			gfx.DrawImage(img1,
						  new RectangleF(0, size.Height*pos, size.Width, size.Height*(1-pos)),
						  new RectangleF(0, img1.Height*pos, img1.Width, img1.Height*(1-pos)),
						  GraphicsUnit.Pixel);
		}

		// draw any partially recalculated replacement image
		if (img2 != null)
			gfx.DrawImage(img2, 0, 0, size.Width, size.Height*pos);
	}
}



// a curve of the form z=f(x,y) shows where a cartesian expression hits zero
class CartesianCurve : SetCurve
{
	public CartesianCurve(ParseTree parseTree)
		: base(parseTree)
		{ }



	// apply an edge detect filter to the image alpha channel
	override protected Bitmap CreateBitmap(int[] data, int w, int h)
	{
		int rgb = data[0] & 0xFFFFFF;

		int[] edges = new int[w*h];

		for (int y=1; y<h-1; y++) {
			for (int x=1; x<w-1; x++) {
				// 3x3 filter kernel
				bool s = Sample(data, x, y, w);

				int c = Compare(s, data, x-1, y-1, w) +
						Compare(s, data, x,   y-1, w) +
						Compare(s, data, x+1, y-1, w) +

						Compare(s, data, x-1, y+1, w) +
						Compare(s, data, x,   y+1, w) +
						Compare(s, data, x+1, y+1, w) +

						Compare(s, data, x-1, y, w) +
						Compare(s, data, x+1, y, w);

				// fake antialiasing
				int a = c*255/4;

				if (s)
					a -= 128;

				a = Util.Clamp(0, a, 255);

				edges[x+y*w] = rgb | (a << 24);
			}
		}

		return base.CreateBitmap(edges, w, h);
	}



	// helper for testing if an alpha pixel is zero
	static bool Sample(int[] data, int x, int y, int w)
	{
		return (data[y*w + x] & 0xFF000000) != 0;
	}



	// helper for testing an edge gradient
	static int Compare(bool s, int[] data, int x, int y, int w)
	{
		return (s != Sample(data, x, y, w)) ? 1 : 0;
	}
}


