// Base class for the watch and plot windows.

using System;
using System.Windows.Forms;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using Microsoft.Win32;



class EricForm : Form
{
	// list of active instances
	static List<Form> mForms = new List<Form>();



	// look up an instance of the requested type
	public static T Instance<T>() where T : class
	{
		foreach (Form form in mForms) {
			T typed = form as T;

			if (typed != null)
				return typed;
		}

		return null;
	}



	// constructor sets default properties
	public EricForm(string title, int x, int y, int w, int h)
	{
		Text = title;

		MinimizeBox = false;
		MaximizeBox = false;

		ShowIcon = false;
		ShowInTaskbar = false;

		DesktopLocation = new Point(RegRead("x", x), RegRead("y", y));
		Size            = new Size (RegRead("w", w), RegRead("h", h));

		StartPosition = FormStartPosition.Manual;

		CheckForIllegalCrossThreadCalls = true;

		mForms.Add(this);
	}



	// display ourselves as a child of the console window
	new public void Show()
	{
		Conio.ConsoleWindow console = new Conio.ConsoleWindow();

		if (!Visible)
			Show(console);

		Activate();

		BringWindowToTop(console.Handle);
	}



	// closing causes comprehensive cleanup
	protected override void OnClosed(EventArgs e)
	{
		mForms.Remove(this);

		// remember the window position
		RegWrite("x", DesktopLocation.X);
		RegWrite("y", DesktopLocation.Y);

		RegWrite("w", Size.Width);
		RegWrite("h", Size.Height);

		// halt background computations, and stop watching for context changes
		mWorkerThread.Stop();

		if (mWatching != null) {
			mWatching.Changed -= Recalculate;
			mWatching = null;
		}

		// send focus somewhere sensible
		if ((mShuttingDown) && (mForms.Count > 0))
			mForms[mForms.Count-1].Activate();
		else
			BringWindowToTop(new Conio.ConsoleWindow().Handle);

		// would the last one to leave please turn out the lights
		if ((mShuttingDown) && (mForms.Count == 0))
			Application.Exit();

		base.OnClosed(e);
	}



	// are we in the process of shutting down?
	static bool mShuttingDown;



	// app is quitting: close all windows
	public static void Shutdown(bool immediate)
	{
		mShuttingDown = true;

		if (immediate) {
			// close everything
			while (mForms.Count > 0)
				mForms[0].Close();
		}
		else if (mForms.Count > 0) {
			// wait for the user to close their windows
			mForms[mForms.Count-1].Activate();

			Application.Run();
		}
	}



	// handle keyboard input
	protected override void OnKeyDown(KeyEventArgs e)
	{
		switch (e.KeyCode) {

			case Keys.Escape:
				// esc to close
				Close();
				e.Handled = true;
				break;

			case Keys.Tab:
				// ctrl+tab to cycle focus
				if (e.Control) {
					CycleFocus();
					e.Handled = true;
				}
				break;
		}

		base.OnKeyDown(e);
	}



	// move focus to the next window in the chain
	public static void CycleFocus()
	{
		if (mShuttingDown)
			return;

		int pos = mForms.IndexOf(ActiveForm) + 1;

		if (pos < mForms.Count)
			mForms[pos].Activate();
		else
			BringWindowToTop(new Conio.ConsoleWindow().Handle);
	}



	// background thread for recalculating the display data
	protected WorkerThread mWorkerThread = new WorkerThread();



	// remember what global variable context we are displaying
	protected Globals mWatching = null;



	// override this to refresh view contents when the context changes
	protected virtual void Recalculate(Globals globals)
	{
		if (globals != mWatching) {
			if (mWatching != null)
				mWatching.Changed -= Recalculate;

			mWatching = globals;
			mWatching.Changed += Recalculate;
		}
	}



	// need a Win32 API to focus the console window
	[DllImport("user32")] static extern bool BringWindowToTop(IntPtr handle);



	// registry key for storing window positions
	const string RegKey = "HKEY_LOCAL_MACHINE\\SOFTWARE\\Shawn Hargreaves\\Eric\\";



	// helper for reading per window registry keys
	int RegRead(string name, int def)
	{
		object o = Registry.GetValue(RegKey+Text, name, null);

		if (o is int)
			return (int)o;

		return def;
	}



	// helper for writing per window registry keys
	void RegWrite(string name, int val)
	{
		Registry.SetValue(RegKey+Text, name, val);
	}
}

