// Window for watching a bunch of expressions, kind of like a spreadsheet.
// You can put values in here, then see how they change as you alter
// constants or redefine functions.

using System;
using System.Windows.Forms;
using System.Collections.Generic;



class Watch : EricForm
{
	// singleton accessor
	public static void Add(List<Function> functions, Globals globals, bool append)
	{
		Watch watch = Instance<Watch>();

		if (watch == null)
			watch = new Watch();

		watch.mView.Add(functions, append);
		watch.Show();
		watch.Recalculate(globals);
	}



	// private constructor
	Watch() : base("Watch", 32, 32, 256, 384)
	{
		mView = new WatchView();

		Controls.Add(mView);

		// grab important keys from the listview back to ourselves
		mView.KeyDown += delegate(object sender, KeyEventArgs e) {
			if ((e.KeyCode == Keys.Escape) || (e.KeyCode == Keys.Tab))
				OnKeyDown(e);
		};

		// tell the listview when we are resized
		Layout += delegate {
			mView.Arrange();
		};
	}



	// the view widget
	WatchView mView;



	// refresh the values of all watched expressions
	protected override void Recalculate(Globals globals)
	{
		base.Recalculate(globals);

		// make copies for safe use by the background thread
		Globals globalsCopy = new Globals(globals);

		List<Function> tasks = new List<Function>();

		foreach (ListViewItem item in mView.Items)
			tasks.Add((Function)item.Tag);

		Dictionary<object, string> results = new Dictionary<object, string>();

		mWorkerThread.BackgroundTask(this,
			// worker thread evaluates a list of expressions
			delegate {
				foreach (Function function in tasks) {
					try {
						results[function] = function.Evaluate(globalsCopy).ToString();
					}
					catch (Exception error) {
						results[function] = error.Message;
					}
				}
			},

			// clear the previous display if this is taking a long time
			delegate {
				foreach (ListViewItem item in mView.Items)
					item.SubItems[WatchView.ValueColumn].Text = "";
			},

			// readback results in the context of the main thread
			delegate {
				foreach (ListViewItem item in mView.Items) {
					string result;

					if (results.TryGetValue(item.Tag, out result))
						item.SubItems[WatchView.ValueColumn].Text = result;
				}
			}
		);
	}



	// display the expressions in a list view widget
	class WatchView : ListView
	{
		public WatchView()
		{
			View = View.Details;
			GridLines = true;
			FullRowSelect = true;
			BorderStyle = BorderStyle.None;
			HeaderStyle = ColumnHeaderStyle.None;

			Columns.Add("pad",      0, HorizontalAlignment.Left);
			Columns.Add("function", 0, HorizontalAlignment.Right);
			Columns.Add("value",    0, HorizontalAlignment.Left);
		}



		// column indices
		public const int ExpressionColumn = 1;
		public const int ValueColumn = 2;



		// add entries to the widget
		public void Add(List<Function> functions, bool append)
		{
			if (!append)
				Items.Clear();

			foreach (Function function in functions) {
				ListViewItem item = new ListViewItem(new string[] { "", function.SourceCode, "" } );

				item.Tag = function;

				Items.Add(item);
				EnsureVisible(Items.Count-1);
			}

			Arrange();
		}



		// update our size and column layout
		public void Arrange()
		{
			BeginUpdate();

			Size = Parent.ClientSize;

			AutoResizeColumn(ExpressionColumn, ColumnHeaderAutoResizeStyle.ColumnContent);

			Columns[ValueColumn].Width = ClientSize.Width - Columns[ExpressionColumn].Width;

			EndUpdate();
		}



		// handle keyboard input
		protected override void OnKeyDown(KeyEventArgs e)
		{
			switch (e.KeyCode) {

				case Keys.A:
					// ctrl+a to select all
					if (e.Control) {
						SelectAll();
						e.Handled = true;
					}
					break;

				case Keys.C:
					// ctrl+c copies to clipboard
					if (e.Control) {
						CopyToClipboard();
						e.Handled = true;
					}
					break;

				case Keys.Delete:
					// delete removes the selection
					RemoveSelection();
					e.Handled = true;
					break;
			}

			base.OnKeyDown(e);
		}



		// selection command handler
		void SelectAll()
		{
			foreach (ListViewItem item in Items)
				item.Selected = true;
		}



		// remove selected items from the control
		void RemoveSelection()
		{
			foreach (ListViewItem item in SelectedItems)
				Items.Remove(item);

			Arrange();
		}



		// copy selected items to the clipboard
		void CopyToClipboard()
		{
			string copy = "";

			foreach (ListViewItem item in SelectedItems)
				copy += item.SubItems[ExpressionColumn].Text + "\t" + item.SubItems[ValueColumn].Text + "\r\n";

			if (copy.Length > 0)
				Clipboard.SetText(copy);
		}
	}
}

