// Interactive editor for reading a line of text. This isn't a proper
// TextReader because it doesn't implement anything other than the
// ReadLine() method, but hey, that's good enough for now.

using System;
using System.Collections.Generic;



class LineEditor : System.IO.TextReader
{
	// input prompts
	const string mPrompt = "\n> ";
	const string mPromptKiller = "\r \r";



	// offset for converting character constants into ASCII control codes
	const int Ctrl = 1 - 'A';



	// interactively input a line of text from the console
	override public string ReadLine()
	{
		Console.Write(mPrompt);

		InputText text = new InputText();

		int historyPos = mHistory.Count;

		for (;;) {
			Conio.Key key = Conio.Read();

			// is it a shortcut hotkey?
			string shortcut;

			if (CheckShortcut(key, text, out shortcut)) {
				if (shortcut != null)
					return shortcut;
				else
					continue;
			}

			switch ((int)key.Unicode) {

				case Ctrl+'A':
				home:
					// home
					text.Pos = 0;
					break;

				case Ctrl+'D':
				del:
					// delete
					if (text.Pos < text.Str.Length)
						text.Str = text.Str.Remove(text.Pos, 1);
					break;

				case Ctrl+'E':
				end:
					// end
					text.Pos = text.Str.Length;
					break;

				case Ctrl+'H':
					// backspace
					if (text.Pos > 0)
						text.Str = text.Str.Remove(--text.Pos, 1);
					break;

				case Ctrl+'K':
					// kill to end of line
					text.Str = text.Str.Substring(0, text.Pos);
					break;

				case Ctrl+'M':
					// accept the current text
					string s = text.Str.Trim();

					if (s.Length > 0) {
						Console.Write("\n");

						mHistory.Remove(s);
						mHistory.Add(s);

						return s;
					}
					break;

				case Ctrl+'[':
					// escape clears the string
					text.Str = "";
					text.Pos = 0;
					break;

				default:
					// insert a character
					if (!char.IsControl(key.Unicode))
						text.Str = text.Str.Insert(text.Pos++, key.Unicode.ToString());
					break;

				case '\0':
					// an extended character code
					switch (key.Keycode) {

						case ConsoleKey.LeftArrow:
							// move left
							if (text.Pos > 0)
								text.Pos--;
							break;

						case ConsoleKey.RightArrow:
							// move right
							if (text.Pos < text.Str.Length)
								text.Pos++;
							break;

						case ConsoleKey.UpArrow:
							// history up
							historyPos = SelectHistory(text, historyPos, -1);
							break;

						case ConsoleKey.DownArrow:
							// history down
							historyPos = SelectHistory(text, historyPos, 1);
							break;

						// reuse existing implementations
						case ConsoleKey.Delete: goto del;
						case ConsoleKey.Home:   goto home;
						case ConsoleKey.End:    goto end;
					}
					break;
			}
		}	
	}



	// input history: oldest at lower indices
	List<string> mHistory = new List<string>();



	// wrapper for accessing the history list
	string History(int n)
	{
		if ((n < 0) || (n >= mHistory.Count))
			return "";

		return mHistory[n];
	}



	// move up or down through the input history buffer
	int SelectHistory(InputText text, int pos, int direction)
	{
		string prefix = text.Str.Substring(0, text.Prev);

		int orig = pos;

		do {
			pos += direction;

			if ((pos < 0) || (pos > mHistory.Count)) {
				if (History(orig).StartsWith(prefix)) {
					pos = orig;
					break;
				}
				else
					return orig;
			}

		} while (!History(pos).StartsWith(prefix));

		text.Str = History(pos);
		text.Pos = text.Str.Length;
		text.Prev = prefix.Length;

		return pos;
	}



	// helper for tracking the current input state, automatically
	// passing text and cursor position changes through to the screen
	class InputText
	{
		string mStr = "";

		int mPos = 0;
		int mPrev = 0;

		int mBase = Console.CursorLeft;



		// current text string
		public string Str
		{
			get {
				return mStr;
			}

			set {
				// echo changes to the screen
				Conio.Write(mBase, Console.CursorTop, value.PadRight(mStr.Length));

				mStr = value;
				mPrev = mPos;
			}
		}



		// cursor position
		public int Pos
		{
			get {
				return mPos;
			}

			set {
				mPos = mPrev = value;

				// echo changes to the screen
				Console.CursorLeft = mBase + mPos;
			}
		}



		// track the previous cursor position for history recall
		public int Prev
		{
			get { return mPrev;  }
			set { mPrev = value; }
		}
	}



	// for checking whether a text shortcut is currently valid
	public delegate bool QueryHandler(string shortcut);

	public event QueryHandler ShortcutQuery;



	// shortcuts can either automate text input, or call one of these
	public delegate void ShortcutHandler();



	// map shortcut keys to either handler delegates or text strings
	Dictionary<ConsoleKeyInfo, object> mShortcuts = new Dictionary<ConsoleKeyInfo, object>();



	// could this be a keyboard shortcut?
	bool CheckShortcut(Conio.Key key, InputText text, out string shortcut)
	{
		shortcut = null;

		object handler;

		if (!mShortcuts.TryGetValue(MakeKeyInfo(key), out handler))
			return false;

		string str = handler as string;

		if (str == null) {
			// call a delegate shortcut
			((ShortcutHandler)handler)();
		}
		else {
			// text shortcuts are only valid when the input string is empty
			if (text.Str.Trim().Length != 0)
				return false;

			// check if this string is currently valid
			if ((ShortcutQuery != null) && (!ShortcutQuery(str)))
				return false;

			// return a text shortcut
			Console.Write(mPromptKiller);

			shortcut = str;
		}

		return true;
	}



	// add a text shortcut
	public void AddShortcut(ConsoleKey key, ConsoleModifiers flags, string text)
	{
		mShortcuts[MakeKeyInfo(key, flags)] = text;
	}



	// add a delegate shortcut
	public void AddShortcut(ConsoleKey key, ConsoleModifiers flags, ShortcutHandler handler)
	{
		mShortcuts[MakeKeyInfo(key, flags)] = handler;
	}



	// helper for constructing a ConsoleKeyInfo struct
	static ConsoleKeyInfo MakeKeyInfo(ConsoleKey key, ConsoleModifiers flags)
	{
		bool shift = (flags & ConsoleModifiers.Shift)   != 0;
		bool alt   = (flags & ConsoleModifiers.Alt)     != 0;
		bool ctrl  = (flags & ConsoleModifiers.Control) != 0;

		return new ConsoleKeyInfo('\0', key, shift, alt, ctrl);
	}



	// helper for constructing a ConsoleKeyInfo struct
	static ConsoleKeyInfo MakeKeyInfo(Conio.Key key)
	{
		bool shift = (key.Flags & Conio.Flags.Shift) != 0;
		bool alt   = (key.Flags & Conio.Flags.Alt)   != 0;
		bool ctrl  = (key.Flags & Conio.Flags.Ctrl)  != 0;

		return new ConsoleKeyInfo('\0', key.Keycode, shift, alt, ctrl);
	}
}
